//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



#ifndef _FPDEFS_H
#define _FPDEFS_H

#define __IEEE_LITTLE_ENDIAN 1

//#ifndef __IEEE_BIG_ENDIAN
//#define __IEEE_BIG_ENDIAN 1
//#endif

#ifndef __IEEE_BIG_ENDIAN
#ifndef __IEEE_LITTLE_ENDIAN
 #error Must define endianness
#endif
#endif

typedef unsigned int __uint32_t ;
typedef int __int32_t ;

/* A union which permits us to convert between a double and two 32 bit
   ints.  */

#ifdef __IEEE_BIG_ENDIAN


typedef union 
{
  double value;
  struct 
  {
    __uint32_t msw;
    __uint32_t lsw;
  } parts;
} ieee_double_shape_type;

#endif

#ifdef __IEEE_LITTLE_ENDIAN

typedef union 
{
  double value;
  struct 
  {
    __uint32_t lsw;
    __uint32_t msw;
  } parts;
} ieee_double_shape_type;

#endif

/* Get two 32 bit ints from a double.  */

#define EXTRACT_WORDS(ix0,ix1,d)                \
do {                                \
  ieee_double_shape_type ew_u;                    \
  ew_u.value = (d);                        \
  (ix0) = ew_u.parts.msw;                    \
  (ix1) = ew_u.parts.lsw;                    \
} while (0)

/* Get the more significant 32 bit int from a double.  */

#define GET_HIGH_WORD(i,d)                    \
do {                                \
  ieee_double_shape_type gh_u;                    \
  gh_u.value = (d);                        \
  (i) = gh_u.parts.msw;                        \
} while (0)

/* Get the less significant 32 bit int from a double.  */

#define GET_LOW_WORD(i,d)                    \
do {                                \
  ieee_double_shape_type gl_u;                    \
  gl_u.value = (d);                        \
  (i) = gl_u.parts.lsw;                        \
} while (0)

/* Set a double from two 32 bit ints.  */

#define INSERT_WORDS(d,ix0,ix1)                    \
do {                                \
  ieee_double_shape_type iw_u;                    \
  iw_u.parts.msw = (ix0);                    \
  iw_u.parts.lsw = (ix1);                    \
  (d) = iw_u.value;                        \
} while (0)

/* Set the more significant 32 bits of a double from an int.  */

#define SET_HIGH_WORD(d,v)                    \
do {                                \
  ieee_double_shape_type sh_u;                    \
  sh_u.value = (d);                        \
  sh_u.parts.msw = (v);                        \
  (d) = sh_u.value;                        \
} while (0)

/* Set the less significant 32 bits of a double from an int.  */

#define SET_LOW_WORD(d,v)                    \
do {                                \
  ieee_double_shape_type sl_u;                    \
  sl_u.value = (d);                        \
  sl_u.parts.lsw = (v);                        \
  (d) = sl_u.value;                        \
} while (0)

double __kernel_sin(double x, double y, int iy);
double __kernel_cos(double x, double y);
double __kernel_tan(double x, double y, int iy);
__int32_t __ieee754_rem_pio2(double x, double *y);
int __kernel_rem_pio2(double *x, double *y, int e0, int nx, int prec, const __int32_t *ipio2);



#include <math.h>


#endif // _FPDEFS_H


