//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: sat3_garbcol.c 393 2007-03-04 23:17:29Z ingo $

#include <hpgcc3.h>
#include <master_globals.h>

extern int __sat3_getblockinfo(SAT_OBJECT ptr,SAT_OBJECT *blkstart,int *blksize);

extern int __sat3_ptrisref(SAT_OBJECT ptr);

extern void __sat3_fixptrs(SAT_OBJECT startptr,SAT_OBJECT endptr,SAT_OBJECT offset);

extern int sat3_free(SAT_OBJECT obj);

// MARK USED AND FREE TEMPOB BLOCKS
// ALSO COLLAPSE ADJACENT FREE BLOCKS INTO ONE


#define SAT3_BLKCACHESIZE 257


// FIND A BLOCK IN TEMPOB QUICK

int __sat3_fastblockinfo(SAT_OBJECT ptr,SAT_OBJECT *blkstart,int *blksize,int *blkcache)
{
int f=(ptr-blkcache[0])*(SAT3_BLKCACHESIZE-1)/(blkcache[SAT3_BLKCACHESIZE-1]-blkcache[0]);
if(f<1) f=0;
else f--;
for(;f<SAT3_BLKCACHESIZE;++f)
{
	if(blkcache[f]<ptr) { --f; break; }
}

if(f==SAT3_BLKCACHESIZE || f<0) return INVALID_PTR;

SAT_OBJECT blk=blkcache[f];
SAT_OBJECT next=0;

while(ptr<blk) {
next=sat3_peek(blk,5);
blk-=next;
}

if(ptr<blk+6) return INVALID_PTR;

*blkstart=blk+6;
*blksize=next-6;
return BLOCK_USED;
}	

void __sat3_initblkcache(int *blkcache)
{
SAT_OBJECT ttop=sat3_peek(SAT_TEMPTOP,5)-5;
SAT_OBJECT tbot=sat3_peek(SAT_TEMPOB,5);
SAT_OBJECT blk=ttop;
SAT_OBJECT next;

int step=(ttop-tbot)/(SAT3_BLKCACHESIZE-1);
int f;
for(f=0;f<SAT3_BLKCACHESIZE-1;++f,++blkcache,ttop-=step)
{
while(ttop<blk) {
next=sat3_peek(blk,5);
blk-=next;
}
*blkcache=blk;	
}
*blkcache=tbot;

}

void __sat3_markblocks(int *blkcache,SAT_OBJECT *firstfree)
{
SAT_OBJECT blk=sat3_peek(SAT_TEMPTOP,5)-5,tbot=sat3_peek(SAT_TEMPOB,5);
SAT_OBJECT next;

do {
sat3_poke(blk+5,0,1);		// MARK AS FREE
next=sat3_peek(blk,5);
blk-=next;
} while(blk>tbot);

sat3_poke(tbot+5,1,1);		// MARK LAST AS USED


// SCAN ALL POINTERS AND MARK USED BLOCKS	

SAT_OBJECT blkend;
int blksize;

SAT_OBJECT tmpptr;
SAT_OBJECT chkptr;
int count;



// CHECK FOR POINTERS IN DATA STACK
//printf("Check stack\n");

tmpptr=sat3_peek(SAT_DSKTOP,5);
while((chkptr=sat3_peek(tmpptr,5))) {
	
if(__sat3_fastblockinfo(chkptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
{
blkend=blk+blksize+5;	
sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
}
tmpptr+=5;
}

// CHECK FOR POINTERS IN RETURN STACK
//printf("Check Rstack\n");

tmpptr=sat3_peek(SAT_RSKTOP,5)-5;
while( (chkptr=sat3_peek(tmpptr,5))& (~0xf)) {
if(__sat3_fastblockinfo(chkptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
{
blkend=blk+blksize+5;	
sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
}	
tmpptr-=5;
}

// CHECK FOR SYSTEM VARIABLES
//printf("Check sysvars\n");

tmpptr=(SAT_OBJECT) SAT_INTRPPTR;

for(count=0x16e;count>=0;--count,tmpptr+=5)
{
chkptr=sat3_peek(tmpptr,5);
if(__sat3_fastblockinfo(chkptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
{
blkend=blk+blksize+5;	
sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
}
}

// CHECK FOR TEMPORARY ENVIRONMENTS
//printf("Check temp. env\n");


tmpptr=sat3_peek(SAT_TEMPENV,5);


do {
	
count=sat3_peek(tmpptr,5);

tmpptr+=10;

if(count>15) {
	count-=15;
	
	do {
	chkptr=sat3_peek(tmpptr,5);
	tmpptr+=5;
	if(__sat3_fastblockinfo(chkptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
	{
	blkend=blk+blksize+5;	
	sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
	}
	chkptr=sat3_peek(tmpptr,5);
	tmpptr+=5;
	if(__sat3_fastblockinfo(chkptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
	{
	blkend=blk+blksize+5;	
	sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
	}
	count-=10;
	}
	while(count>0);
	
	count=1;	// ENSURE LOOP IS REPEATED
	continue; 
}
	
} while(count>0);


// CHECK FOR PTRS IN THE TOUCHTAB TABLE
//printf("Check touchtab\n");

tmpptr=sat3_peek(SAT_TOUCHTAB,5)+3;

while( (chkptr=sat3_peek(tmpptr,5))) {
if(__sat3_fastblockinfo(chkptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
{
blkend=blk+blksize+5;	
sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
}
tmpptr+=5;
}

// CHECK FOR PTRS IN THE FSTVGERPTR TABLE
//printf("Check vger\n");

tmpptr=sat3_peek(SAT_FSTVGERPTR,5);


if(tmpptr) {
	// CHECK FOR THE FSTVGERPTR ITSELF
	if(__sat3_fastblockinfo(tmpptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
	{
	blkend=blk+blksize+5;	
	sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
	}
	
	tmpptr+=15;
	count=sat3_peek(tmpptr,5);
	while(count--) {
		tmpptr+=20;
		while( (chkptr=sat3_peek(tmpptr,5))) {
			if(__sat3_fastblockinfo(chkptr,&blk,&blksize,blkcache)!=INVALID_PTR) 
			{
			blkend=blk+blksize+5;	
			sat3_poke(blkend,1,1);	// MARK BLOCK AS REFERENCED
			}
			tmpptr+=5;
		}
	}
	

}

// COLLAPSE ADJACENT FREE BLOCKS

blk=sat3_peek(SAT_TEMPTOP,5)-5;
//printf("Collapsing free blocks\n");

SAT_OBJECT nextfree=0;
do {
next=sat3_peek(blk,5);
if(next) {
if(sat3_peek(blk+5,1)==0) {  // CURRENT BLOCK IS FREE
	if(sat3_peek(blk-next+5,1)==0)	{	// FOUND CONSECUTIVE FREE BLOCKS
	sat3_poke(blk,next+sat3_peek(blk-next,5),5);	// COLLAPSE BLOCKS
	}
	else {	// NEXT BLOCK IS USED, STORE NEXTFREE INFO
	if(nextfree) sat3_poke(blk-5,nextfree-blk,5);
	else sat3_poke(blk-5,0,5);
	nextfree=blk;
	blk-=next;
	
	}
}
else { blk-=next; }
}
} while(next);


*firstfree=nextfree;


}




void sat3_garbcol()
{
    // IMPLEMENT A GARBAGE COLLECTION ON TEMPOB
    // ALL UNREFERENCED OBJECTS ARE REMOVED
 
	SAT_OBJECT blk,ffree,stblk,endblk,nextfree;
	int blksize,objsize,tmpsize;
	int *blkcache;


	blkcache=(int *)malloc(SAT3_BLKCACHESIZE*sizeof(int));
	if(!blkcache) return;
	
	// INITIALIZE TEMPOB CACHE
	__sat3_initblkcache(blkcache);
	
	//printf("Mark free blocks...");
	__sat3_markblocks(blkcache,&ffree);	// MARK ALL USED AND FREE BLOCKS IN TEMPOB
	//printf("DONE\n");
	
	
	//printf("Adjusting free blocks...");
	
	if(ffree>0) {	// AT LEAST ONE FREE BLOCK IN CHAIN
	while(sat3_peek(ffree-5,5)) {
	nextfree=ffree+sat3_peek(ffree-5,5);
	stblk=ffree+6;		// START OF BLOCK TO MOVE
	endblk=nextfree-sat3_peek(nextfree,5)+6;  // END OF BLOCK TO MOVE
	blk=ffree-sat3_peek(ffree,5)+6;  // DESTINATION ADDRESS
	
	// ADJUST POINTERS AND MOVE MEMORY
	__sat3_fixptrs(stblk,endblk,blk-stblk);
	sat3_memmove(blk,stblk,endblk-stblk);	
	
	sat3_poke(nextfree,nextfree-(blk-6)-(endblk-stblk),5);
	
	ffree=nextfree;
	} 
	
	
	// COLLAPSE LAST BLOCK	
	stblk=ffree+6;		// START OF BLOCK TO MOVE
	endblk=sat3_peek(SAT_RSKTOP,5);  // END OF BLOCK TO MOVE
	blk=ffree-sat3_peek(ffree,5)+6;  // DESTINATION ADDRESS
	
	// ADJUST POINTERS AND MOVE MEMORY
	__sat3_fixptrs(stblk,endblk,blk-stblk);
	sat3_memmove(blk,stblk,endblk-stblk);	
	
	sat3_poke(SAT_RSKTOP,endblk+(blk-stblk),5);	// FIX RSKTOP POINTER
   	// TEMPTOP
   	sat3_poke(SAT_TEMPTOP,sat3_peek(SAT_TEMPTOP,5)+(blk-stblk),5);
    
	
	}
	//printf("DONE\n");

	// CHECK FOR BAD SIZED OBJECTS
	//printf("Check for size...");	
	
	blk=sat3_peek(SAT_TEMPTOP,5)+1;

	//printf("Start block loop\n");
	while((blksize=sat3_peek(blk-6,5)))
	{
			blk-=blksize;
			objsize=0;
			while(objsize<blksize) {
				tmpsize=sat3_sizeof(blk+objsize);
				if(tmpsize<=5) break;	
				objsize+=tmpsize;
			}
			objsize+=6;
			if(blksize!=objsize) {
				// OVERSIZED COMPOSITE OR INVALID_OBJECT
				if(blksize>objsize && objsize>11) {
					// VALID OBJECT W/OVERSIZED BLOCK, SHRINK
					// PATCH BLOCK SIZE
					
					//printf("Resizing=%05X,new=%05X\n",blksize,objsize);
					sat3_poke(blk+objsize-6,objsize,5);
					
					stblk=blk+blksize;
					endblk=sat3_peek(SAT_RSKTOP,5);  // END OF BLOCK TO MOVE
					
					// ADJUST POINTERS AND MOVE MEMORY
					__sat3_fixptrs(stblk,endblk,blk+objsize-stblk);
					sat3_memmove(blk+objsize,stblk,endblk-stblk);	
	
					sat3_poke(SAT_RSKTOP,endblk+(blk+objsize-stblk),5);	// FIX RSKTOP POINTER
   					// TEMPTOP
   					sat3_poke(SAT_TEMPTOP,sat3_peek(SAT_TEMPTOP,5)+(blk+objsize-stblk),5);				
				}
				else return; // INVALID REFERENCED OBJECT, CRASH IS IMMINENT
			}
	}
	
				    // ADJUST FREE MEMORY
				    
    				int freemem=sat3_peek(SAT_DSKTOP,5)-sat3_peek(SAT_RSKTOP,5);
    				freemem/=5;
    				sat3_poke(SAT_AVMEM,freemem,5);
			
//printf("Done GC\n");
 // CALCULATOR'S GARBAGE COLLECTION ALSO CLEAN THE VIRTUAL STACK
 // NOT IMPLEMENTED HERE
}


